class SolrError(Exception):

    def __init__(self, message=None, status_code=None):
        if isinstance(message, tuple):
            message = message[1]

        super(SolrError, self).__init__(message)
        self.status_code = status_code


class SolrSchemaUpdateError(SolrError):

    def __init__(self, fields, message=None, status_code=None):
        self.fields = [field["name"] for field in fields]
        if message is None:
            message = "Unable to update SOLR schema for %s" % ", ".join(
                self.fields
            )
        super(SolrSchemaUpdateError, self).__init__(
            message=message,
            status_code=status_code
        )


class SolrSchemaValidationError(SolrError):

    def __init__(self, field, message=None):
        self.field = field
        if message is None:
            message = "Field (%s) is not in SOLR schema fields" % field
        super(SolrSchemaValidationError, self).__init__(message=message)


class SolrDuplicateUniqueKeyError(SolrError):
    def __init__(self, pk):
        self.pk = pk
        message = "The unique key %s already exists" % self.pk
        super(SolrDuplicateUniqueKeyError, self).__init__(message=message)


class SolrDocumentNotExistError(SolrError):
    def __init__(self, pk):
        self.pk = pk
        message = "The document with unique key %s does not exist" % self.pk
        super(SolrDocumentNotExistError, self).__init__(message=message)


class SolrDeleteUniqueKeyError(SolrError):
    def __init__(self, pk):
        self.pk = pk
        message = "The unique key %s can not be deleted" % self.pk
        super(SolrDeleteUniqueKeyError, self).__init__(message=message)


class SolrUnspecifiedOperatorError(SolrError):
    def __init__(self, field_name):
        self.field_name = field_name
        message = "The operator is not specified in %s" % self.field_name
        super(SolrUnspecifiedOperatorError, self).__init__(message=message)


class SolrUnspportedOperatorError(SolrError):
    def __init__(self, operator):
        self.operator = operator
        message = "The operator:%s is not supported" % self.operator
        super(SolrUnspportedOperatorError, self).__init__(message=message)



import wukong.errors as solr_errors
import copy
import re
import six

INFINITE_ROWS = 999999999


class SolrNode(object):
    """
    The base class to model a tree node in the query logic to SOLR
    """
    def __init__(self, *args, **kwargs):
        self.items = self.build_items(args, kwargs)

    @property
    def parsed_solr_query(self):
        """
        Parse the current node to a query string
        """
        pass

    @classmethod
    def build_items(cls, args, kwargs):
        """
        Build a list of items under the current logic operator
        """
        items = []
        for arg in args:
            items.append(arg)

        for key, value in kwargs.items():
            try:
                name, operator = key.split("__")
            except ValueError:
                raise solr_errors.SolrUnspecifiedOperatorError(key)

            items.append(Comparator(operator, name, value))
        return items


class AND(SolrNode):
    """
    Model the AND logic in SOLR query
    """
    logic = "AND"

    @property
    def parsed_solr_query(self):
        if len(self.items) == 0:
            return ""
        if len(self.items) == 1:
            return self.items[0].parsed_solr_query
        return "(%s)" % (" %s " % self.logic).join([
            node.parsed_solr_query for node in self.items
        ])


class OR(SolrNode):
    """
    Model the OR logic in SOLR query
    """
    logic = "OR"

    @property
    def parsed_solr_query(self):
        if len(self.items) == 0:
            return ""
        if len(self.items) == 1:
            return self.items[0].parsed_solr_query
        return "(%s)" % (" %s " % self.logic).join([
            node.parsed_solr_query for node in self.items
        ])


class NOT(SolrNode):
    """
    Model the NOT logic in SOLR query
    """
    logic = "NOT"

    def __init__(self, *args, **kwargs):
        super(NOT, self).__init__(*args, **kwargs)

        if len(self.items) > 1:
            raise solr_errors.SolrError(
                "Logic node:NOT only supports one operand"
            )

    @property
    def parsed_solr_query(self):
        if len(self.items) == 0:
            return ""
        return "!(%s)" % self.items[0].parsed_solr_query


class Comparator(SolrNode):
    """
    Model the compare logic in SOLR query
    """
    logic = "COMP"

    def __init__(self, operator, key, value):
        self.operator = operator
        self.key = key
        self.value = value

    @property
    def parsed_solr_query(self):
        query_string = ""

        if self.key is None:
            return query_string

        operator = self.operator
        key = self.key
        value = self.value

        # EQUAL
        if operator == 'eq':
            if value is None:
                query_string = "-%s:[* TO *]" % key
            elif isinstance(value, six.string_types):
                query_string = "%s:\"%s\"" % (key, value)
            else:
                query_string = "%s:%s" % (key, value)

        # NOT EQUAL
        elif operator == 'ne':
            if value is None:
                query_string = "%s:*" % key
            elif isinstance(value, six.string_types):
                query_string += "-%s:\"%s\"" % (key, value)
            else:
                query_string += "-%s:%s" % (key, value)

        # IN
        elif operator == 'in':
            if not isinstance(value, list):
                raise solr_errors.SolrError(
                    "Operator:in only takes list type"
                )
            new_value = []
            for i in value:
                if i is not None:
                    if isinstance(i, six.string_types):
                        new_value.append("\"%s\"" % i)
                    else:
                        new_value.append(str(i))
            list_str = "(%s)" % " ".join(new_value)

            query_string = "%s:%s" % (key, list_str)

        # NOT IN
        elif operator == 'nin':
            if not isinstance(value, list):
                raise solr_errors.SolrError(
                    "Operator:not in only takes list type"
                )
            new_value = []
            for i in value:
                if i is not None:
                    if isinstance(i, six.string_types):
                        new_value.append("\"%s\"" % i)
                    else:
                        new_value.append(str(i))
            list_str = "(%s)" % " ".join(new_value)

            query_string = "-%s:%s" % (key, list_str)

        # WILD CARD
        elif operator == 'wc':
            value = value or ""
            if '*' in value:
                query_string = "%s:%s" % (key, value)
            else:
                query_string = "%s:*%s*" % (key, value)
            query_string = query_string.replace("**", "*")

        # NOT WILD CARD
        elif operator == 'nwc':
            value = value or ""
            if '*' in value:
                query_string = "-%s:%s" % (key, value)
            else:
                query_string = "-%s:*%s*" % (key, value)
            query_string = query_string.replace("**", "*")

        # GREATER THAN
        elif operator == 'g':
            query_string = "%s:{%s TO *]" % (key, value or "*")

        # GREATER THAN OR EQUAL TO
        elif operator == 'ge':
            query_string = "%s:[%s TO *]" % (key, value or "*")

        # LESS THAN
        elif operator == 'l':
            query_string = "%s:[* TO %s}" % (key, value or "*")

        # LESS THAN OR EQUAL TO
        elif operator == 'le':
            query_string = "%s:[* TO %s]" % (key, value or "*")

        # EXISTS
        elif operator == 'ex':
            query_string = "%s:[* TO *]" % key

        # NOT EXISTS
        elif operator == 'nex':
            query_string = "-%s:[* TO *]" % key

        # UNKNOWN
        else:
            raise solr_errors.SolrUnspportedOperatorError(operator)

        return query_string


class SolrQueryManager(object):
    """
    A class to chain different query methods for SOLR
    and construct the final query to SOLR
    """
    def __init__(
        self,
        doc_class,
        node=None,
        sort_str=None,
        weight_dict=None,
        returned_fields=None,
        edismax=False,
        rows=INFINITE_ROWS,
        start=0,
        facet_fields=None,
        facet_options={},
        mincount=1,
        facet_group=True,
        group_fields=None,
        group_limit=0,
        group_options={},
        boost_func=None,
        bf_weight=1,
        boost_query=None,
        bq_weight=1,
        minimum_matches=None,
        text_keywords=None,
        stats_fields=None,
    ):

        self.node = node if node else AND()
        self.sort_str = sort_str
        self.weight_dict = weight_dict
        self.returned_fields = returned_fields
        self.edismax = edismax
        self.start = start
        self.rows = rows
        self.doc_class = doc_class
        self.facet_fields = facet_fields
        self.facet_options = facet_options
        self.mincount = mincount
        self.facet_group = facet_group
        self.group_fields = group_fields
        self.group_limit = group_limit
        self.group_options = group_options
        self.boost_func = boost_func
        self.bf_weight = bf_weight
        self.boost_query = boost_query
        self.bq_weight = bq_weight
        self.minimum_matches = minimum_matches
        self.text_keywords = text_keywords
        self.stats_fields = stats_fields

    def to_dict(self):
        """
        Get the json representation of the query manager
        """
        return {
            "node": self.node,
            "doc_class": self.doc_class,
            "sort_str": self.sort_str,
            "weight_dict": self.weight_dict,
            "returned_fields": self.returned_fields,
            "edismax": self.edismax,
            "start": self.start,
            "rows": self.rows,
            "facet_fields": self.facet_fields,
            "facet_options": self.facet_options,
            "mincount": self.mincount,
            "facet_group": self.facet_group,
            "group_fields": self.group_fields,
            "group_limit": self.group_limit,
            "group_options": self.group_options,
            "boost_func": self.boost_func,
            "bf_weight": self.bf_weight,
            "boost_query": self.boost_query,
            "bq_weight": self.bq_weight,
            "minimum_matches": self.minimum_matches,
            "text_keywords": self.text_keywords,
            "stats_fields": self.stats_fields,
        }

    def filter(self, *args, **kwargs):
        """
        Filter the SOLR documents by mathmatical and logical operators.
        Usage:
            filter(name__eq="Test Name")
            filter(name__eq="Test Name", city__wc="Test*")
            filter(OR(name__eq="Test Name", city__wc="Test*"),
                    population__ge=300000)
        """
        node = copy.deepcopy(self.node)
        items = node.items if hasattr(node, 'items') else []

        if node.logic != "AND":
            node = AND(node)
            items = node.items

        items += SolrNode.build_items(args, kwargs)
        if len(items) == 1:
            node = copy.deepcopy(items[0])

        params = self.to_dict()
        params.update({
            "node": node
        })
        return SolrQueryManager(**params)

    def sort_by(self, sort_str):
        """
        Sort the SOLR documents
        Usage:
            ascending: sort_by("name")
            descending: sort_by("-date")
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "sort_str": sort_str
        })
        return SolrQueryManager(**params)

    def group_by(self, group_fields, group_limit=0, **kwargs):
        """
        Group the SOLR documents
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "group_fields": group_fields,
            "group_limit": group_limit,
            "group_options": kwargs
        })
        return SolrQueryManager(**params)

    def facet(self, facet_fields, mincount=1, group=True, **kwargs):
        """
        Facet the SOLR documents
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "facet_fields": facet_fields,
            "mincount": mincount,
            "facet_group": group,
            "facet_options": kwargs
        })
        return SolrQueryManager(**params)

    def stats(self, stats_fields):
        node = copy.deepcopy(self.node)
        params = self.to_dict()

        params.update({
            'node': node,
            'stats_fields': stats_fields,
        })
        return SolrQueryManager(**params)

    def boost_by_func(self, boost_func, bf_weight=1):
        """
        Boost query by function
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "boost_func": boost_func,
            "bf_weight": bf_weight
        })
        return SolrQueryManager(**params)

    def boost_by_query(self, boost_query, bq_weight=1):
        """
        Boost query by query
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "boost_query": boost_query,
            "bq_weight": bq_weight
        })
        return SolrQueryManager(**params)

    def search(self, text, minimum_matches=None, **weights):
        """
        Search SOLR by text query.
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "edismax": True,
            "text_keywords": text,
            "minimum_matches": minimum_matches,
            "weight_dict": weights
        })
        return SolrQueryManager(**params)

    def only(self, *args):
        """
        Specify the returned fields in each document.
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "returned_fields": args
        })
        return SolrQueryManager(**params)

    def limit(self, rows):
        """
        Specify the number of documents returned.
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "rows": rows
        })
        return SolrQueryManager(**params)

    def offset(self, start):
        """
        Specify the offset of the entire documents
        """
        node = copy.deepcopy(self.node)
        params = self.to_dict()
        params.update({
            "node": node,
            "start": start
        })
        return SolrQueryManager(**params)

    @property
    def query(self):
        """
        Construct the query string to SOLR
        """
        solr_query = self.node.parsed_solr_query

        if solr_query == "":
            solr_query = "*:*"

        params = {
            "rows": self.rows,
            "start": self.start,
            "wt": "json"
        }

        if self.edismax:
            params['defType'] = 'edismax'
            params["fq"] = solr_query
            text = self.text_keywords
            text = text.strip()
            query_string = " ".join([k for k in re.compile("\s+").split(text)])
            params["q"] = query_string
        else:
            params["q"] = solr_query

        if self.weight_dict:
            params["qf"] = " ".join(
                ["%s^%s" % (k, self.weight_dict[k]) for k in self.weight_dict]
            )

        if self.minimum_matches:
            params["mm"] = self.minimum_matches

        if self.returned_fields:
            params['fl'] = ','.join(self.returned_fields)

        if self.sort_str:
            splits = self.sort_str.split("-", 1)
            if len(splits) > 1:
                sort_field = splits[1]
                sort_dir = "desc"
            else:
                sort_field = splits[0]
                sort_dir = "asc"

            params['sort'] = "%s %s" % (sort_field, sort_dir)

        if self.group_fields:
            params['group'] = "on"
            params['group.ngroups'] = "true"
            params['group.field'] = self.group_fields
            params['group.limit'] = self.group_limit
            params.update(dict([
                ("group.{}".format(k), v) for k, v in
                self.group_options.items()
            ]))

        if self.facet_fields:
            params['facet'] = "on"
            params['facet.field'] = self.facet_fields
            params['facet.mincount'] = self.mincount
            params.update(dict([
                ("facet.{}".format(k), v) for k, v in
                self.facet_options.items()
            ]))

        if self.stats_fields:
            params['stats'] = "true"
            params['stats.field'] = self.stats_fields

        if self.boost_func:
            params['bf'] = "%s^%s" % (self.boost_func, self.bf_weight)

        if self.boost_query:
            params['bq'] = "%s^%s" % (self.boost_query, self.bq_weight)

        return params

    def get(self, *args, **kwargs):
        """
        Fetch one document from SOLR

        :return: document from SOLR
        :rtype: SolrDoc
        """
        return self.filter(*args, **kwargs).one()

    def create(self, field_weights=None, **kwargs):
        """
        Create one document in SOLR

        :return: document created in SOLR
        :rtype: SolrDoc
        """
        doc = self.doc_class(
            field_weights=field_weights,
            **kwargs
        )

        kwargs = {"%s__eq" % doc.unique_key: doc.get_unique_field()}
        if self.get(**kwargs):
            raise solr_errors.SolrDuplicateUniqueKeyError(
                doc.get_unique_field()
            )
        doc.index(commit=True)
        return doc

    def update(self, field_weights=None, **kwargs):
        """
        Update one document in SOLR

        :return: document created in SOLR
        :rtype: SolrDoc
        """
        doc = self.doc_class(
            partial_update=True,
            field_weights=field_weights,
            **kwargs
        )
        kwargs = {"%s__eq" % doc.unique_key: doc.get_unique_field()}
        solr_doc = self.get(**kwargs)
        if solr_doc is None or solr_doc.get_unique_field() != \
                doc.get_unique_field():
            raise solr_errors.SolrDocumentNotExistError(doc.get_unique_field())
        doc.index(commit=True)
        return doc

    def raw(self, **extra):
        """
        Retrieve matched documents from SOLR in json format

        :return: documents from SOLR
        :rtype: list of json
        """
        return self.doc_class.solr.select(self.query, **extra)

    def groups(self, **extra):
        """
        Retrieve document groups when group is ON in the query

        :return: document groups of SOLR documents
        :rtype: dict
        """
        result = self.raw(groups=True, **extra)

        return result['groups']

    def facets(self, **extra):
        """
        Retrieve document facets when facet is ON in the query

        :return: facet counts of SOLR documents
        :rtype: dict
        """
        result = self.raw(facets=True, **extra)

        return result['facets']

    def all(self, **extra):
        """
        Retrieve all matched documents from SOLR and convert them into SolrDocs

        :return: documents from SOLR
        :rtype: list of SolrDoc
        """
        from wukong.models import SolrDocs

        result = self.raw(**extra)

        return SolrDocs(docs=self.doc_class.from_json_docs(result['docs']))

    def one(self, **extra):
        """
        Get one document from SOLR and convert it into SolrDoc

        :return: one document from SOLR
        :rtype: SolrDoc
        """
        new_query_manager = self.limit(1)

        result = new_query_manager.raw(**extra)

        if len(result['docs']) == 0:
            return None

        return self.doc_class(**result['docs'][0])



import kazoo.client
from kazoo.exceptions import NoNodeError
import json
from collections import defaultdict
import itertools

import logging
logger = logging.getLogger(__name__)


def _get_hosts_from_state(state):
    """
    Given a SOLR state json blob, extract the active hosts

    :param state dict: SOLR state blob
    :returns: list[str]
    """
    active_nodes = set()
    for shard, shard_data in state.get('shards', {}).items():
        replicas = shard_data['replicas']
        for replica, replica_data in replicas.items():
            if replica_data['state'] == 'active':
                node_url = replica_data['base_url'][:-5]  # strip /solr
                node_url = node_url.replace('http://', '')
                active_nodes.add(node_url)

    return active_nodes


def _zk_data_to_dict(data):
    """json load data retreived with zk_client.get()"""
    if not isinstance(data, str):
        data = data.decode('utf-8')
    return json.loads(data)


class Zookeeper(object):
    """
    Retrieve the status of SOLR servers from Zookeeper
    """
    def __init__(self, hosts):
        self.hosts = hosts

    def _get_active_hosts(self):
        active_hosts = defaultdict(set)

        try:
            zk_client = kazoo.client.KazooClient(
                hosts=self.hosts,
                read_only=True
            )
            zk_client.start(timeout=5)
        except Exception:
            return active_hosts

        try:
            collections = zk_client.get_children('/collections')
        except NoNodeError:
            # No collections have been created on the zookeeper host yet.
            logger.debug('No collections found')
            collections = []

        # Fetch any potential collection aliases from solr
        aliases = {}
        try:
            aliases = _zk_data_to_dict(zk_client.get('/aliases.json')[0])
            if aliases is None:  # If there are no aliases, we'll get None
                aliases = {}
        except:
            logger.debug('No /aliases.json file found')

        # Handle SOLR 6+ style state.json paths
        for collection in collections:
            try:
                state_path = '/collections/{}/state.json'.format(collection)
                state_json = zk_client.get(state_path)[0]
            except NoNodeError:
                logger.debug(
                    'No SOLR 6 state found for collection [%s]',
                    collection
                )
            else:
                state = _zk_data_to_dict(state_json)
                active_hosts[collection] |= _get_hosts_from_state(
                    state.get(collection, {})
                )

        # Handle SOLR <6 style clusterstate.json
        try:
            cluster_state_str = zk_client.get('/clusterstate.json')[0]
        except Exception:
            cluster_state_str = '{}'

        cluster_state = _zk_data_to_dict(cluster_state_str)

        for collection_name, state in cluster_state.items():
            hosts = _get_hosts_from_state(state)
            active_hosts[collection_name] |= hosts

        logger.error('Got aliases: %s', aliases)
        for alias_name, member_string in aliases.get('collection', {}).items():
            members = member_string.split(',')

            # Only support single member aliases for now, because we need to
            # be able to fetch schemas.
            member = members[0]

            active_hosts[alias_name] = active_hosts[member].copy()

        zk_client.stop()

        return active_hosts

    def get_active_hosts(self, collection_name=None):
        """
        Get the current active SOLR hosts from Zookeeper

        :param collection_name: If provided, the name of a SOLR collection to
                                get the hosts for. If not provided, all solr
                                hosts will be returned.
                                Optional.

        :returns list[str]: A list of solr nodes in the form `http://hostname`
        """
        logger.debug('Getting active hosts for collection %s', collection_name)
        active_hosts = self._get_active_hosts()

        if collection_name is not None:
            return list(active_hosts.get(collection_name, []))
        else:
            return list(itertools.chain.from_iterable(
                self._get_active_hosts().values()
            ))



import logging
import datetime as dt
import wukong.errors as solr_errors
from wukong.request import SolrRequest
from wukong.zookeeper import Zookeeper
import json

logger = logging.getLogger(__name__)


class SolrAPI(object):

    def __init__(self, solr_hosts, solr_collection,
                 zookeeper_hosts=None, timeout=15):
        """
        Do all the interactions with SOLR server
        (e.g. update, select, get and delete)

        :param solr_hosts: the hosts for SOLR.
        :type server: str

        :param solr_collection: the name of the collection in SOLR.
        :type solr_collection: str

        :param zookeeper_hosts: the hosts for zookeeper.
        :type zookeeper_hosts: str

        :param timeout: the timeout for request to SOLR.
        :type timeout: int

        """

        if solr_hosts is None and zookeeper_hosts is not None:
            logger.info(
                'Getting solr hosts from zookeeper for collection %s',
                solr_collection
            )
            zk = Zookeeper(zookeeper_hosts)
            solr_hosts = zk.get_active_hosts(collection_name=solr_collection)

        if solr_hosts is None or solr_collection is None:
            logger.error('Neither solr_hosts nor solr_collection has been set')
            raise solr_errors.SolrError(
                "Either solr_hosts or solr_collection can not be None"
            )

        if not isinstance(solr_hosts, list):
            solr_hosts = solr_hosts.split(",")

        if zookeeper_hosts is not None:
            hostnames, sep, chroot = zookeeper_hosts.rpartition('/')

            # If hostnames is empty then there is no chroot. Set it to empty.
            if not hostnames:
                chroot = ''
            else:
                chroot = '/%s' % chroot

            logger.debug('Using solr via zookeeper at chroot %s', chroot)

            self.zookeeper_hosts = [
                "http://%s%s" % (host, chroot,)
                for host in zookeeper_hosts.split(",")
            ]

            logger.info(
                'Connected to zookeeper hosts at %s',
                self.zookeeper_hosts
            )

        else:
            logger.debug('Not using zookeeper for SolrCloud')
            self.zookeeper_hosts = None

        logger.info('Connected to solr hosts %s', solr_hosts)
        self.solr_hosts = ["http://%s/solr/" % host for host in solr_hosts]

        self.solr_collection = solr_collection

        self.client = SolrRequest(
            solr_hosts=self.solr_hosts,
            zookeeper_hosts=zookeeper_hosts,
            timeout=timeout
        )

    def _get_collection_url(self, path):
        return "%s/%s" % (self.solr_collection, path)

    def is_alive(self):
        """
        Check if current collection is live from zookeeper.

        :return: weather or not if the collection is live
        :rtype: boolean
        """
        params = {'detail': 'true', 'path': '/clusterstate.json'}

        try:
            response = self.client.get('zookeeper', params)
        except solr_errors.SolrError:
            logger.exception('Failed to check zookeeper')
            return False
        else:
            try:
                data = json.loads(response['znode']['data'])
            except ValueError:
                return False

            for name, collection in data.items():
                shards = collection['shards']
                for shard, shard_info in shards.items():
                    replicas = shard_info['replicas']
                    for replica, info in replicas.items():
                        state = info['state']
                        if name == self.solr_collection and state != 'active':
                            return False

            return True

    def update(self, docs, commit=False):
        """
        Add new docs or updating existing docs.

        :param docs: a list of instances of SolrDoc.
        :type server: list

        :param commit: whether or not we should commit the documents.
        :type server: boolean

        """
        if not docs:
            return

        data = json.dumps(
            docs,
            default=lambda obj: obj.isoformat() if isinstance(
                obj, dt.datetime) else None
        )

        params = {}

        if commit:
            params['commit'] = 'true'

        return self.client.post(
            self._get_collection_url('update/json'),
            params=params,
            body=data
        )

    def select(self,
               query_dict,
               groups=False,
               facets=False,
               stats=False,
               **kwargs
               ):
        """
        Query documents from SOLR.

        :param query_dict: a dict containing the query params to SOLR
        :type query_dict: dict

        :param metadata: whether or not solr metadata should be returned
        :type metadata: boolean

        :param kwargs: a dict of additional params for SOLR
        :type kwargs: dict

        :return: reformatted response from SOLR
        :rtype: dict
        """

        if kwargs:
            query_dict.update(kwargs)

        response = self.client.get(
            self._get_collection_url('select'),
            params=query_dict
        )

        data = {}
        if groups and 'grouped' in response:
            data['groups'] = response['grouped']

        if facets and 'facet_counts' in response:
            data['facets'] = response['facet_counts']

        if stats and 'stats' in response:
            data['stats'] = response['stats']

        if 'response' in response and 'docs' in response['response']:
            response_data = response['response']
            data['docs'] = response_data['docs']
            data['total'] = response_data.get('numFound', len(data['docs']))

        return data

    def delete(self, unique_key, unique_key_value, commit=False):
        """
        Deleting a document from SOLR.

        :param unique_key: the unique key for the doc to delete
        :param unique_key_value: the value for the unique_key
        :param commit: whether or not we should commit the documents.
        :type server: boolean

        """
        params = {}

        if commit:
            params['commit'] = 'true'

        data = json.dumps({"delete": {"query": "%s:%s" %
                                      (unique_key, unique_key_value)}})

        return self.client.post(
            self._get_collection_url('update/json'),
            params=params,
            body=data
        )

    def commit(self):
        """
        Hard commit documents to SOLR.
        """
        params = {'commit': 'true'}

        return self.client.post(
            self._get_collection_url('update/json'), params=params)

    def get_schema(self):
        """
        Get the SOLR schema for the solr collection.

        :return: the schema for the current collection
        :rtype: dict
        """
        response = self.client.get(self._get_collection_url('schema'))

        return response.get('schema', {})

    def add_schema_fields(self, fields):
        """
        Add new fields to the schema of current collection

        :param fields: a list of dicts of fields.
        :type fields: list

        """
        if not fields:
            return

        data = json.dumps(fields)

        try:
            return self.client.post(
                self._get_collection_url('schema/fields'),
                body=data
            )
        except solr_errors.SolrError as e:
            raise solr_errors.SolrSchemaUpdateError(fields, message=e.args[0])



from wukong.api import SolrAPI
from wukong.query import SolrQueryManager
import wukong.errors as solr_errors
from six import with_metaclass
import re


class SolrDocMetaClass(type):
    """
    Meta class for SolrDoc
    """

    @property
    def solr(self):
        """
        Return a instance of SOLR api class.
        """
        if not hasattr(self, '_solr'):
            self._solr = SolrAPI(
                solr_hosts=self.solr_hosts,
                solr_collection=self.collection_name,
                zookeeper_hosts=self.zookeeper_hosts,
                timeout=self.request_timeout
            )
        return self._solr

    @property
    def documents(self):
        """
        Return a instance of SOLR Query Manager class.
        """
        return SolrQueryManager(doc_class=self)

    @property
    def unique_key(self):
        """
        Return the unique key of the collection schema
        """
        return self.schema.get("uniqueKey")

    @property
    def schema(self):
        """
        Return the current collection schema
        """
        if not hasattr(self, '_schema'):
            self._schema = self.solr.get_schema()
        return self._schema


class SolrDocs(object):
    """
    A Wrapper Container for a collection of SolrDoc
    instances, designed for batch update and delete
    """

    def __init__(self, docs=None):
        self.docs = []

        self.doc_class = None
        for doc in docs or []:
            self.add(doc)

    def __str__(self):
        return str(self.docs)

    def __repr__(self):
        return '%s(len(%s)=%s)' % (
            self.__class__.__name__,
            self.doc_class.__name__ if self.doc_class else "doc",
            len(self.docs))

    def __setitem__(self, key, value):
        self.docs[key] = value

    def __getitem__(self, key):
        return self.docs[key]

    def __len__(self):
        return len(self.docs)

    def __iter__(self):
        for doc in self.docs:
            yield doc

    def add(self, doc):
        """
        Add a document to the SolrDoc container

        :param doc: the document to add
        :type doc: SolrDoc
        """
        if self.doc_class:
            if doc.__class__ != self.doc_class:
                raise solr_errors.SolrError(
                    "The types of documents in a " +
                    "container should be the same"
                )
        else:
            self.doc_class = doc.__class__

        self.docs.append(doc)

    def index(self, commit=False):
        """
        Index all current documents in the container to SOLR collection

        :param commit: whether or not to commit to SOLR when submitted
        :type commit: boolean
        """
        if len(self.docs) == 0:
            return

        data = []

        for doc in self.docs:
            data.append(doc.get_data_for_solr())

        self.doc_class.solr.update(data)

        if commit:
            self.doc_class.solr.commit()

    def delete(self, commit=False):
        """
        Index all current documents in the container to SOLR collection

        :param commit: whether or not to commit to SOLR when submitted
        :type commit: boolean
        """
        if len(self.docs) == 0:
            return

        unique_keys_str = "(%s)" % ' '.join(
            [str(doc.get_unique_field()) for doc in self.docs]
        )
        self.doc_class.solr.delete(
            self.doc_class.unique_key,
            unique_keys_str,
            commit=commit)


class SolrDoc(with_metaclass(SolrDocMetaClass, object)):
    """
    The base class for modeling any type of solr document
    """
    solr_hosts = None
    zookeeper_hosts = None
    collection_name = None
    request_timeout = 15

    @property
    def solr(self):
        """
        Reference the property from the metaclass and
        return a instance of SOLR api class.
        """
        return self.__class__.solr

    @property
    def documents(self):
        """
        Reference the property from the metaclass and
        return a instance of SOLR Query Manager class.
        """
        return self.__class__.documents

    @property
    def unique_key(self):
        """
        Reference the property from the metaclass and
        return the unique key of the collection schema
        """
        return self.__class__.unique_key

    @property
    def schema(self):
        """
        Reference the property from the metaclass and
        return the current collection schema
        """
        return self.__class__.schema

    def __init__(self, partial_update=False, field_weights=None, **kwargs):
        """
        :param partial_update: whether or not only the provided
            fields should be updated, otherwise, the entire
            document will be overridden.
        :type partial_update: boolean

        :param field_weights: the weights for different fields for boosting
        :type field_weights: dict

        :param kwargs: the dictionary of key value mapping
            for fields of the document
        :param kwargs: dict
        """

        self.fields = {}

        self.partial_update = partial_update

        self.field_weights = field_weights

        self.set_fields(**kwargs)

        self.validate_schema_fields(self.fields)

    @classmethod
    def add_schema_fields(cls, fields):
        """
        Add fields to the SOLR schema which will hit the SOLR schema api.

        :param fields: a list of field meta info (e.g. name, type)
        :type fields: list
        """
        schema_fields = [field["name"] for field in cls.schema['fields']]

        added_fields = []
        for field in fields:
            if field["name"] not in schema_fields:
                added_fields.append(field)

        return cls.solr.add_schema_fields(added_fields)

    @classmethod
    def from_json_docs(cls, json_docs):
        """
        Convert a list of json dict from SOLR to a list of SolrDoc.

        :param json_docs: a list of dict returned from SOLR server
        :type json_docs: list

        :return: a list of SolrDoc
        :rtype: list
        """
        docs = []
        for doc in json_docs:
            docs.append(cls(**doc))

        return docs

    def __str__(self):
        return str(self.fields)

    def __repr__(self):
        return '%s(%s=%s)' % (
            self.__class__.__name__, self.unique_key,
            self.get_unique_field()
        )

    def __eq__(self, other):
        """
        Compare two SOLR documents

        :param other: the other doc to compare
        :type other: SolrDoc

        :return: whether or not the fields of the two docs are the same
        :rtype: boolean
        """

        return self.fields == other.fields

    def __ne__(self, other):
        """
        Compare two SOLR documents

        :param other: the other doc to compare
        :type other: SolrDoc

        :return: whether or not the fields of the two docs are different
        :rtype: boolean
        """

        return self.fields != other.fields

    def __setitem__(self, key, value):
        self.set_field(key, value)

    def __getitem__(self, key):
        return self.get_field(key)

    def __getattr__(self, key):
        item = self.get_field(key)
        return item

    def __len__(self):
        return len(self.fields)

    def __iter__(self):
        for field in self.fields:
            yield field, self.get_field(field)

    def get_unique_field(self):
        """
        Get the value of the unique key in the SolrDoc
        """
        return self.fields.get(self.unique_key, None)

    def set_fields(self, **fields):
        """
        Set the values for all fields in the SolrDoc
        """
        for field in fields:
            # if the data is from SOLR, we don't want '_version_'
            if fields[field] is not None and field != '_version_':
                self.set_field(field, fields[field])

    def get_field(self, key):
        """
        Get the value of an field from the SolrDoc
        """
        return self.fields.get(key, None)

    def set_field(self, key, value):
        """
        Set the value of an field from the SolrDoc
        """
        self.fields[key] = value

    def del_field(self, key):
        """
        Delete the value of an field from the SolrDoc
        """
        if key in self.fields:
            if key == self.unique_key:
                raise solr_errors.SolrDeleteUniqueKeyError(key)
            del self.fields[key]

    def is_partial_update(self):
        """
        Return if the SolrDoc is only for partial update
        """
        return self.partial_update

    def set_partial_update(self, value):
        """
        Set the value of partial update
        """
        self.partial_update = bool(value)

    def set_field_weight(self, field, weight):
        """
        Set the weight of an field from the SolrDoc
        """
        if self.field_weights is None:
            self.field_weights = {}

        self.field_weights[field] = weight

    def get_field_weight(self, field):
        """
        Get the weight of an field from a SolrDoc
        """
        if self.field_weights is None:
            return None

        return self.field_weights.get(field)

    @classmethod
    def validate_schema_fields(cls, fields):
        """
        Validate if the fields are valid for SOLR by checking with the schema

        :param fields: a list of dicts for the field meta info
        :type fields: list

        :return: whether or not the fields are consistent with SOLR schema
        :rtype: boolean
        """
        schema_fields = dict([(field["name"], field) for field
                              in cls.schema.get("fields", {})])

        dynamic_fields = dict([(field["name"], field) for field
                               in cls.schema.get("dynamicFields", {})])

        pk = cls.unique_key

        if pk not in fields or fields[pk] is None:
            raise solr_errors.SolrSchemaValidationError(
                pk,
                message="Unique key is not specified"
            )

        for key, field in fields.items():
            if key in schema_fields:
                if schema_fields[key].get("multiValued") and \
                        not isinstance(field, list):
                    raise solr_errors.SolrSchemaValidationError(
                        key,
                        message="%s is not list type" % key
                    )
            else:
                found = False

                for df in dynamic_fields:
                    df_re = "^%s$" % df.replace("*", ".*")
                    if re.match(df_re, key):
                        found = True

                if not found:
                    raise solr_errors.SolrSchemaValidationError(key)

    def get_data_for_solr(self):
        """
        Generate the data for SOLR indexing for the current SolrDoc.

        :return: a json representation of a SolrDoc for SOLR indexing.
            If the SolrDoc is invalid, it returns None.
        :rtype: dict
        """
        self.validate_schema_fields(self.fields)

        solr_data = {}

        for key in self.fields:
            value = self.get_field(key)
            if self.is_partial_update() and key != self.unique_key:
                # {"set": self.get_attr(key)} is for partial update in SOLR
                # if "set" is not used, the entire entity in SOLR will be
                # replaced
                value = {"set": value}

            # boost is to adjust the weight of a specific field for
            # that SolrDoc
            if self.get_field_weight(key):
                value = {
                    "value": value,
                    "boost": self.get_field_weight(key)
                }

            solr_data[key] = value

        return solr_data

    def index(self, commit=False):
        """
        Index the current SolrDoc from SOLR.

        :param commit: whether or not to commit upon submission.
        :type commit: boolean
        """
        solr_doc = self.get_data_for_solr()

        self.solr.update([solr_doc])
        if commit:
            self.solr.commit()

    def delete(self, commit=False):
        """
        Delete the current SolrDoc from SOLR.

        :param commit: whether or not to commit upon submission.
        :type commit: boolean
        """
        unique_key = self.unique_key
        unique_key_value = self.get_unique_field()
        self.solr.delete(unique_key, unique_key_value, commit=commit)



import os
from setuptools import setup, find_packages


def read(fname):
    return open(os.path.join(os.path.dirname(__file__), fname)).read()

try:
    import pypandoc
    README = pypandoc.convert('README.md', 'rst')
    CHANGES = pypandoc.convert('CHANGES.md', 'rst')
except:
    README = read('README.md')
    CHANGES = read('CHANGES.md')

version = '0.3.1'

setup(
    name="wukong",
    version=version,
    author="Da Kuang",
    author_email="dkuang1980@gmail.com",
    description="An ORM Client library for SolrCloud",
    keywords="solr cloud solrcloud client python orm",
    url="https://github.com/SurveyMonkey/wukong",
    long_description='%s\n\n%s' % (README, CHANGES),
    install_requires=[
        "requests",
        "kazoo",
        "six>=1.6.1"
    ],
    tests_require=read('test-requirements.txt'),
    packages=find_packages(exclude=['tests']),
    classifiers=[
        "Topic :: Software Development :: Libraries :: Python Modules",
        "License :: OSI Approved :: MIT License",
        "Natural Language :: English",
        "Operating System :: OS Independent",
        "Programming Language :: Python",
    ],
)



from wukong.zookeeper import Zookeeper
from requests.exceptions import ConnectionError
from wukong.errors import SolrError
import requests
import random
import json
import time

try:
    from urlparse import urljoin
except:
    from urllib.parse import urljoin


def process_response(response):
    if response.status_code != 200:
        raise SolrError(response.reason)
    try:
        response_content = json.loads(response.text)
    except:
        raise SolrError("Parsing Error: %s" % response.text)

    return response_content


class SolrRequest(object):
    """
    Handle requests to SOLR and response from SOLR
    """
    def __init__(self, solr_hosts, zookeeper_hosts=None, timeout=15):
        self.client = requests.Session()
        self.master_hosts = solr_hosts
        self.current_hosts = solr_hosts
        self.servers = []
        self.timeout = timeout
        if zookeeper_hosts is not None:
            self.zookeeper = Zookeeper(zookeeper_hosts)
        else:
            self.zookeeper = None

        self.last_error = None
        # time to revert to old host list (in minutes) after an error
        self.check_hosts = 5

    def request(self, path, params, method, body=None):
        """
        Prepare data and send request to SOLR servers
        """
        def handle_error():
            # if the sdk knows where zookeeper lives
            if self.zookeeper is not None:
                self.last_error = time.time()
                # get the current list of active nodes from zookeeper
                curr_hosts = self.zookeeper.get_active_hosts()
                if len(curr_hosts) == 0:
                    # all nodes are down - so raise an error
                    raise SolrError("SOLR reporting all nodes as down")
                else:
                    # if all nodes are not down, update the current_hosts list
                    # and the servers list
                    self.current_hosts = curr_hosts
                    self.servers = curr_hosts

            host = self.servers.pop(0)
            return make_request(host, path)

        headers = {'content-type': 'application/json'}
        extraparams = {'wt': 'json',
                       'omitHeader': 'true',
                       'json.nl': 'map'}

        if params is None:
            params = {}

        params.update(extraparams)

        # if there hasn't been an error in 5 minutes, reset the solr_hosts
        if self.last_error is not None and \
                (time.time() - self.last_error) % 60 > self.check_hosts:

            self.current_hosts = self.master_hosts
            self.last_error = None

        def make_request(host, path):
            fullpath = urljoin(host, path)
            try:
                response = self.client.request(
                    method,
                    fullpath,
                    params=params,
                    headers=headers,
                    data=body,
                    timeout=self.timeout
                )

                # Connected to the node, but didn't get a successful response
                if (len(self.servers) > 0 and
                        hasattr(response, 'status_code') and
                        response.status_code != 200):
                    # try with another node
                    handle_error()

                return response

            # Didn't successfully connect to the node
            except ConnectionError as e:
                if len(self.servers) > 0:
                    # try with another node
                    handle_error()

                raise SolrError(str(e))

        self.servers = list(self.current_hosts)
        if len(self.servers) == 0:
            handle_error()

        random.shuffle(self.servers)
        host = self.servers.pop(0)

        response = make_request(host, path)
        return process_response(response)

    def post(self, path, params=None, body=None):
        """
        Send a POST request to the SOLR servers
        """
        return self.request(path, params, 'POST', body=body)

    def get(self, path, params=None):
        """
        Send a GET request to the SOLR servers
        """
        return self.request(path, params, 'GET')
